#!/usr/bin/env python
################################################################################
# PAV Pipeline: compare corresponding K-mers from two K-mer files. Output two
# files each containing the K-mers present only in each of the data sets.
# Note: jellyfish count outputs expected. Each line should be in the form of:
# >K-mer occurrence
#
# Try inputting the smaller file first which will save memory.
#
# Hong Ching Lee     30 June 2011
################################################################################

# import libraries
import argparse # parse input options
#import math # mathematical operations


def find_PAKs(in_name_1, in_name_2, out_name_1, out_name_2, min_occ, is_verbose=False):
    '''This method finds the Presence / Absence K-mers (PAKs) between two input
    files. A K-mer m is defined as a PAK if:
    1. m is present in one data set (D1) but not in the other (D2) and occ(m) >=
        min_occ.'''

    # store first file into a dict with the K-mers as keys and occ as objects
    mers_1 = {}

    with open(in_name_1, "rU") as in_1:
        for line in in_1:
            columns = line.split()
            mers_1[columns[0]] = int(columns[1])
    in_1.close()

    out_1 = open(out_name_1, "w")
    out_2 = open(out_name_2, "w")
    with open(in_name_2, "rU") as in_2:
        for line in in_2:
            columns = line.split()
            k_mer = columns[0]
            occ = int(columns[1])

            if (k_mer not in mers_1):
                if (occ >= min_occ):
                    out_2.write(k_mer + "\t" + str(occ) + "\n")
            else:
                # delete this k_mer from mers_1 to avoid re-print
                del mers_1[k_mer]
    in_2.close()

    # write the remaining mers_1 items to out_1...those with >= min_occ anyway
    for k_mer in mers_1:
        if mers_1[k_mer] >= min_occ:
            out_1.write(k_mer + "\t" + str(mers_1[k_mer]) + "\n")

    out_1.close()
    out_2.close()


def check_param(parser, args):
    '''This method checks the input parameters.'''

    if (args.in_1 is None):
        print("===Please specify input file 1.===")
        parser.print_help()
    elif (args.in_2 is None):
        print("===Please specify input file 2.===")
        parser.print_help()
    elif (args.out_1 is None):
        print("===Please specify output file 1.===")
        parser.print_help()
    elif (args.out_2 is None):
        print("===Please specify output file 2.===")
        parser.print_help()
    else:
        return True

    return False


if __name__ == '__main__':

    # set input options
    usage = "%(prog)s --in1 <input_file_1> --in2 <input_file_2> --out1 <output_file_1> --out2 <output_file_2> [-m <min_occ_cutoff>]\n\n"
    descriptions = "For PAV Pipeline: find Presence / Absence K-mers (PAKs).\nNote: jellyfish count outputs expected. Each line should be in the form of:\n>K-mer occurrence"
    parser = argparse.ArgumentParser(usage=usage, description=descriptions)
    parser.add_argument("--in1", dest="in_1", required=True, help="input file 1 name")
    parser.add_argument("--in2", dest="in_2", required=True, help="input file 2 name")
    parser.add_argument("--out1", dest="out_1", required=True, help="output file 1 name")
    parser.add_argument("--out2", dest="out_2", required=True, help="output file 2 name")
    parser.add_argument("-m", "--min_occ", dest="min_occ", help="Occurrences below this cutoff are discarded [default=%(default)s]", type=int, default=4)
    parser.add_argument("-v", "--verbose", dest="is_verbose", action="store_true", help="be verbose [default=%(default)s]", default=False)

    args = parser.parse_args()

    if check_param(parser, args):
        find_PAKs(args.in_1, args.in_2, args.out_1, args.out_2, args.min_occ, args.is_verbose)
