#!/usr/bin/env python

################################################################################
# PAV Pipeline: compare two samples of sequencing reads and find Presence /
# Absence Variations (PAVs)
#
# Hong Ching Lee     22 December 2011
# Philipp E. Bayer, 2014: various bugfixes, now calls SOAPaligner, compares
# against gff3 files etc.
################################################################################

# import libraries
try:
    import ConfigParser # parse the config file
except ImportError: # this happens with python 3
    import configparser as ConfigParser

import argparse # parse input options
import os
import sys
from multiprocessing import Pool
import functools
import itertools
import subprocess
import math
import glob
import logging
import time
from Bio import SeqIO

VERSION = 1.5

def execute_commands(commands_list, ncpus, is_verbose):
    if ncpus == 1:
        for i in range(len(commands_list)):
            uncurry_execute_command((commands_list[i], is_verbose))
    else: # multiprocessing
        pool = Pool(processes=ncpus)
        # Note: this is a work-around of the functools.partial non-picklable
        # issue. Otherwise functools.partial should be used instead: e.g.
        # execute_command_partial = functools.partial(execute_command, is_verbose=is_verbose)
        # pool.imap_unordered(execute_command_partial, commands_list)
        pool.imap_unordered(uncurry_execute_command, itertools.izip(commands_list, itertools.repeat(is_verbose)))
        pool.close()
        pool.join()


def uncurry_execute_command(params): #cmd, is_verbose):
    if params[1]: # basically if is_verbose
        print(params[0])

    try:
        p = subprocess.Popen(params[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = p.communicate()
        return_code = p.wait()
        logging.info('Executing %s', params[0])

        if return_code != 0:
            sys.stderr.write("Problems executing command: " + params[0] + '\n')
            sys.stderr.write(err + "\n")
            sys.exit(1)

    except OSError:
        msg = "Execution of command \"" + cmd + "\" failed: " + err
        logging.critical(msg)
        sys.stderr.write(msg + "\n" + params[0] + "\nStopping the pipeline.\n")
        sys.exit(1)

def clean_up_files(file_list, is_verbose):
    '''Goes through a list of file paths and remove them.'''

    for filename in file_list:
        if is_verbose:
            print("Removing file:", filename)

        try:
            os.remove(filename)
            logging.info(filename + " removed.")
        except OSError:
            msg = "Failed to clean file: \"" + filename
            logging.critical(msg)
            sys.stderr.write(msg + '\n')


def execute_jellyfish(config, config_filename, sample_1, sample_2, out_dir, kmer_size, ncpus, s1_jellyfish_name, s2_jellyfish_name, is_verbose):
    '''Executes jellyfish count on a specific k-mer size.'''

    print_message("Executing Jellyfish...")

    base_cmd = "python " + config.get('PIPELINE', 'SCRIPTS_PATH') +\
        "/jellyfish_spawn.py -o " + out_dir + " -c " + config_filename +\
        " -k " + str(kmer_size) + " -n " + str(ncpus) + " -m dump -s "
    uncurry_execute_command((base_cmd + " ".join(sample_1) + " -t " + s1_jellyfish_name, is_verbose))
    uncurry_execute_command((base_cmd + " ".join(sample_2) + " -t " + s2_jellyfish_name, is_verbose))


def execute_split_jellyfish(pipeline_path, in_file1_name, in_file2_name, n_split_bases, ncpus, is_verbose):
    '''Executes the split_kmers.py script.'''

    print_message("Splitting the jellyfish outputs into subsets...")
    cmd1 = "python " + pipeline_path + "/split_kmers.py -i " + in_file1_name + " -n " + str(n_split_bases)
    cmd2 = "python " + pipeline_path + "/split_kmers.py -i " + in_file2_name + " -n " + str(n_split_bases)

    execute_commands([cmd1, cmd2], ncpus, is_verbose)
    clean_up_files([in_file1_name, in_file2_name], is_verbose)


def execute_find_PAKs(pipeline_path, in_file1_name, in_file2_name, n_splits, min_occ, ncpus, is_verbose):
    '''Executes the find_PAKs.py script.'''

    print_message("Finding Presence / Absence K-mers (PAKs) from jellyfish outputs...")
    out1_basename = in_file1_name + "_paks"
    out2_basename = in_file2_name + "_paks"
    commands = []
    in_files = []

    for i in range(n_splits):
        in1_name = in_file1_name + "." + str(i)
        in2_name = in_file2_name + "." + str(i)
        out1_name = out1_basename + "." + str(i)
        out2_name = out2_basename + "." + str(i)
        in_files.append(in1_name)
        in_files.append(in2_name)

        cmd = "python " + pipeline_path + "/find_PAKs.py --in1 " + in1_name + " --in2 " + in2_name + " --out1 " + out1_name + " --out2 " + out2_name + " -m " + str(min_occ)
        commands.append(cmd)

    execute_commands(commands, ncpus, is_verbose)
    clean_up_files(in_files, is_verbose)


def execute_cat_PAKs(out_dir, s1_jellyfish_name, s2_jellyfish_name, ncpus, is_verbose):
    '''Concatenates the split PAKs files into one per sample.'''

    print_message("Concatenating the split PAKs files into a single file per sample...")
    cmd_s1 = "cat " + out_dir + "/kmers/" + s1_jellyfish_name + "_paks.* > " + out_dir + "/kmers/" + s1_jellyfish_name + "_paks"
    cmd_s2 = "cat " + out_dir + "/kmers/" + s2_jellyfish_name + "_paks.* > " + out_dir + "/kmers/" + s2_jellyfish_name + "_paks"

    execute_commands([cmd_s1, cmd_s2], ncpus, is_verbose)

    # clean up the split pak files
    s1_split_pak_files = glob.glob(s1_jellyfish_name + "_paks.*")
    s2_split_pak_files = glob.glob(s2_jellyfish_name + "_paks.*")
    clean_up_files(s1_split_pak_files, is_verbose)
    clean_up_files(s2_split_pak_files, is_verbose)


def execute_find_PARs(pipeline_path, sample_1, sample_2, s1_jellyfish_name, s2_jellyfish_name, mode, out_dir, ncpus, is_verbose):
    '''Execute the find_PARs.py script.'''

    print_message("Identifying the Presence / Absence Reads (PARs)...")

    commands = []
    s1_paks_file_name = out_dir + "/kmers/" + s1_jellyfish_name + "_paks"
    s2_paks_file_name = out_dir + "/kmers/" + s2_jellyfish_name + "_paks"

    for i in range(len(sample_1) / 2):
        reads_A_file_name = sample_1[i * 2]
        reads_B_file_name = sample_1[(i * 2) + 1]
        output_header = out_dir + "/PARs/s1_" + str(i) + "_PARs"
        cmd = "python " + pipeline_path + "/find_PARs.py -k " + s1_paks_file_name +\
            " -A " + reads_A_file_name + " -B " + reads_B_file_name + " -o " +\
            output_header + " -m " + mode
        commands.append(cmd)

    for i in range(len(sample_2) / 2):
        reads_A_file_name = sample_2[i * 2]
        reads_B_file_name = sample_2[(i * 2) + 1]
        output_header = out_dir + "/PARs/s2_" + str(i) + "_PARs"
        cmd = "python " + pipeline_path + "/find_PARs.py -k " + s2_paks_file_name +\
            " -A " + reads_A_file_name + " -B " + reads_B_file_name + " -o " +\
            output_header + " -m " + mode
        commands.append(cmd)

    execute_commands(commands, ncpus, is_verbose)


def generate_out_dirs(out_dir):
    '''Given the output directory path, creates sub-directories for storing
    outputs from each stage of the pipeline.'''
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    if not os.path.isdir(out_dir + "/kmers"):
        os.makedirs(out_dir + "/kmers")
    if not os.path.isdir(out_dir + "/PARs"):
        os.makedirs(out_dir + "/PARs")
    if not os.path.isdir(out_dir + "/SOAPs"):
        os.makedirs(out_dir + "/SOAPs")
    if not os.path.isdir(out_dir + "/Results"):
        os.makedirs(out_dir + "/Results")

def setup_logger(out_dir, is_verbose):
    '''Given the output directory path, sets up a logger.'''
    # set up time in UNIX time, as integer
    time_str = str(int(time.time()))
    log_filename = out_dir + '/pipeline_' + time_str + '.log'

    if is_verbose:
        logging.basicConfig(filename=log_filename, format='%(asctime)s %(levelname)s: %(message)s', datefmt='%Y-%b-%d %H:%M:%S', level=logging.DEBUG)
    else:
        logging.basicConfig(filename=log_filename, format='%(asctime)s %(message)s', level=logging.INFO)
    print_message("Logging to %s" %log_filename)


def print_message(msg):
    '''Formats and prints a simple message'''
    print("\n====================")
    print(msg)
    print("====================\n")
    logging.info(msg)


def check_params(parser, args):
    if args.ncpus <= 0:
        sys.stderr.write("\nNumber of CPUs must be a positive integer\n")
    elif (len(args.sample_1) % 2) or (len(args.sample_2) % 2):
        sys.stderr.write("\nInput FASTA files must be in pairs. E.g. --s1 sample1_A.fasta sample1_B.fasta...\n")
    else:
        return True

    return False


def check_files(first_file, second_file):
   ''' Loads three lines from FASTA files, first and third has to start with >'''
   print_message("Checking input files...")
   for file in [first_file, second_file]:
       for sub_file in file:
           fh = open(sub_file)
           first = fh.readline()
           fh.readline()
           second = fh.readline()
           if first[0] != ">" and second[0] != ">":
               sys.stderr.write("Error: Given FASTA files are not of format of one line per identifier, one line per sequence\n")
               sys.exit(1)

def execute_soap(out_dir, reference_path, ncpus, is_verbose):
    commands = []
    if not reference_path.endswith(".index"):
        reference_path += ".index"
    for element in os.listdir(os.path.join( out_dir, "PARs")):
        element = os.path.join(out_dir, "PARs", element)
        if element.endswith("single.fasta") and os.path.getsize(element):
             out_paired = element.replace(".fasta", "_paired.soap").replace("PARs", "SOAPs")
             out_single = element.replace(".fasta", "_single.soap").replace("PARs", "SOAPs")
             cmd = "soap -D %s -2 %s -o %s -a %s -r 2 -M 0" %(reference_path, out_single, out_paired, element)
             commands.append(cmd)
    execute_commands(commands, ncpus, is_verbose)

def compare_soap_to_gff(out_dir, gff_path):
    ''' Parses SOAP output files, compares them to the gff in the given gff-path '''

    # Parse the gene positions into a dictionary (there are always less genes than SOAP-aligned reads)
    genes = open(gff_path)
    known_genes_dict = {}
    names_dict = {} # key: tuple (chrom, start, end), value: gene-name
    lines_dict = {} # keeps for each gene name the entire line from the gff
    
    duplicate_c = 0 # counts how many duplicate genes there are in the gff
    for line in genes:
      ll = line.rstrip('\n\r').split('\t')
      chrom, start, end, name = ll[0], int(ll[3]), int(ll[4]), ll[-1]
      if (chrom, start, end) in known_genes_dict:
        duplicate_c += 1
      known_genes_dict[(chrom, start,end)] = ll
      names_dict[(chrom, start,end)] = name
      lines_dict[name] = line
 
    print_message('Got %s genes' %len(known_genes_dict))

    output_dir = os.path.join(out_dir, "Results")
    # this one contains the rest
    soap_dir = os.path.join(out_dir, "SOAPs")
    pars_dir = soap_dir.replace('SOAPs', 'PARs')

    # one output file per SOAP file
    
    out_file = open(os.path.join(output_dir, "gff_compared_to_soap_results.txt"), "w")
    gene_file = open(os.path.join(output_dir, "gene_names_hit.txt"), 'w')
    missing_gene_file = open(os.path.join(output_dir, 'missing_gene_names.txt'), 'w')

    print_message('Writing results file to %s, %s and %s' %(out_file.name, gene_file.name, missing_gene_file.name))
    out_file.write("Name\tNumber_of_total_genes\tDuplicate_genes\tNumber_of_genes_hit_by_reads\tTotal_reads\tAligned_reads\tComplete_hit\tOnly_start\tOnly_end\tNumber_alignments_without_hit_to_gene\tDuplicate_hits\n")
    seen_soaps = set()
    for element in os.listdir(soap_dir):
        seen_soaps.add(os.path.join(soap_dir, element))

        fasta = os.path.join(pars_dir, element.replace('_paired.soap', '.fasta').replace('_single.soap', '.fasta').replace('SOAPs', 'PARs'))
        assert os.path.exists(fasta)
        # since we have only two lines per read, we can do the following
        total_reads = int(os.popen('wc -l %s' %fasta).read().split(' ')[0])
        assert total_reads % 2 == 0, 'FATAL ERROR: There is an uneven number of lines in "%s". Is the file a Fasta file with two lines per record?'%fasta
        total_reads /= 2

        soap_fh = open(os.path.join(soap_dir, element))
        seen_names = set() # keeps the names of READS
        seen_genes = set() # keeps the names of GENES that are hit by reads

        # Various counters
        complete_c = 0 # all reads that hit completely inside genes
        only_start_c = 0 # reads where only start hits
        only_end_c = 0 # reads where only end hits
        no_align_counter = 0 # reads with no alignment in a gene
        duplicate_hits = 0 # how often a read appears more than once
        line_c = 0 # how many reads there are in the soap file

        print soap_fh.name
        for line in soap_fh:
            ll = line.split('\t')
            try:
                name, chrom, start, length = ll[0], ll[7], int(ll[8]), int(ll[5])
            except ValueError:
                sys.stderr.write('FATAL ERROR: Something went wrong while parsing the given gff3 file. Please check whether the file has a format similar to the example gff.\n')
                sys.exit(1)
            end = start + length
            line_c += 1
            if name in seen_names:
                duplicate_hits += 1
            aligns = False
            # Now check the PAR against all known genes
            for gene in known_genes_dict:
                gene_chrom, gene_start, gene_end = gene
                if chrom != gene_chrom: continue
                # hit completey inside gene
                if (start >= gene_start and end <= gene_end):
                    complete_c += 1
                    aligns = True
                    gene_name = names_dict[gene]
                    seen_genes.add(gene_name)
                    break
                else:
                    # only hit_start inside gene
                    if (start >= gene_start and start < gene_end):
                        only_start_c += 1
                        aligns = True
                        gene_name = names_dict[gene]
                        seen_genes.add(gene_name)
                        break
                  # only hit_end inside gene
                    if (end <= gene_end and end > gene_start):
                        only_end_c += 1
                        aligns = True
                        gene_name = names_dict[gene]
                        seen_genes.add(gene_name)
                        break
            if not aligns:
                no_align_counter += 1
            seen_names.add(name)

        missing_genes = set(names_dict.values()) - seen_genes

        full_name = os.path.join(soap_dir, element.replace('_paired.soap', '').replace('_single.soap', ''))
        line = full_name + "\t"  + str(len(known_genes_dict)) + '\t' + str(duplicate_c) + '\t' + str(len(seen_genes)) + '\t' + str(total_reads) + "\t" + str(line_c) + "\t" + str(complete_c) + "\t" + str(only_start_c) + "\t" + str(only_end_c) + "\t" + str(no_align_counter) + "\t" + str(duplicate_hits) + '\n'
        out_file.write(line)
        
        gene_file.write('#%s\n'%full_name)
        for element in seen_genes:
          gene_file.write(lines_dict[element])
        missing_gene_file.write('#%s\n'%full_name)
        for element in missing_genes:
          missing_gene_file.write(lines_dict[element])
    
    # check for non-existing soap-files
    for soap in seen_soaps:
        if "s1" in soap:
            other_soap = soap.replace("s1", "s2")
        else:
            other_soap = soap.replace("s2", "s1")
        if other_soap not in seen_soaps:
            name = other_soap.replace('_paired.soap', '').replace('_single.soap', '')
            out_file.write(name + "\t-\t-\t-\t-\t-\t-\t-\t-\n")
         
        

def get_best_kmer(identify_kmer_path, identify_kmer_conf):
    cmd = "perl " + identify_kmer_path + " " + identify_kmer_conf
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out, err = p.communicate()
    exit_code = p.wait()
    if int(exit_code) != 0:
        sys.stderr.write("Error: %s" %err)
        sys.exit(1)
    kmer = int(out.split()[-1])
    message = []
    if "The optimalKmerSize file exist " in err:
        message.append("WARNING: old identify_kmer results exist, re-using these.")
        # could use safe_eval here - but: one more dependency!
        path_to_delete = err.split("\n")[2]
        path_to_delete = path_to_delete.split("[")[1].split("]")[0]
        message.append('If you want to re-run the k-mer identification, delete "%s".' %path_to_delete)
    message.append("Best kmer is: %s" %kmer)
    print_message("\n".join(message))
    return kmer

if __name__ == '__main__':

    # set input options
    descriptions = "PAV Pipeline: compare two samples of sequencing reads and find Presence / Absence Reads."
    parser = argparse.ArgumentParser(description=descriptions)
    parser.add_argument("--s1", dest="sample_1", nargs='+', required=True, help="input sample 1 paired reads fasta files")
    parser.add_argument("--s2", dest="sample_2", nargs='+', required=True, help="input sample 2 paired reads fasta files")
    parser.add_argument("-o", "--out_dir", dest="out_dir", help="output directory", required=True)
    parser.add_argument("-c", "--config_file", dest="config", required=True, help="input config file")
    parser.add_argument("-n", "--ncpus", dest="ncpus", help="input the number of CPUs to use [default=%(default)s]", type=int, default=1)
    parser.add_argument("-m", "--mode", dest="mode", help="Method for identifying PARs: all-paks or all-bases [default: %(default)s]", choices=["all-paks", "all-bases"], default="all-paks")
    parser.add_argument("-v", "--verbose", dest="is_verbose", action="store_true", help="be verbose [default=%(default)s]")
    parser.add_argument("-V", "--version", action="version", help='prints version number and exits', version='Version: %s' %VERSION)

    args = parser.parse_args()

    if not check_params(parser, args):
        sys.exit(1)

    config = ConfigParser.ConfigParser()
    config.read(args.config)

    generate_out_dirs(args.out_dir)
    setup_logger(args.out_dir, args.is_verbose)

    # parse config file
    pipeline_path = config.get('PIPELINE', 'SCRIPTS_PATH')
    n_split_bases = config.getint('ADVANCED', 'N_SPLIT_BASES')

    run_soap = True
    run_gff = True
    try:
        reference_path = config.get('PIPELINE', 'REFERENCE_PATH')
    except ConfigParser.NoOptionError:
        logging.info("No reference given, skipping SOAP alignment and gff3 comparison")
        # skip the soap run later
        run_soap = False
        run_gff = False

    try:
        gff_path = config.get('PIPELINE', 'GFF_PATH')
    except ConfigParser.NoOptionError:
        logging.info("No gff given, skipping gff3 comparison")
        run_gff = False
 
    n_split_files = int(math.pow(5, n_split_bases))
    min_occ = config.getint('ADVANCED', 'MIN_OCC')
    s1_jellyfish_name = "s1_jellyfish"
    s2_jellyfish_name = "s2_jellyfish"
    kmers_dir = args.out_dir + "/kmers/"
    in_file1_name = kmers_dir + s1_jellyfish_name
    in_file2_name = kmers_dir + s2_jellyfish_name

    # run the actual pipeline
    check_files(args.sample_1, args.sample_2)
    identify_kmer_path = os.path.join(pipeline_path, 'identifyKmerSize.pl')
    kmer = get_best_kmer(identify_kmer_path, args.config)

    #execute_jellyfish(config, args.config, args.sample_1, args.sample_2, kmers_dir, kmer, args.ncpus, s1_jellyfish_name, s2_jellyfish_name, args.is_verbose)
    #execute_split_jellyfish(pipeline_path, in_file1_name, in_file2_name, n_split_bases, args.ncpus, args.is_verbose)
    #execute_find_PAKs(pipeline_path, in_file1_name, in_file2_name, n_split_files, min_occ, args.ncpus, args.is_verbose)
    #execute_cat_PAKs(args.out_dir, s1_jellyfish_name, s2_jellyfish_name, args.ncpus, args.is_verbose)
    #execute_find_PARs(pipeline_path, args.sample_1, args.sample_2, s1_jellyfish_name, s2_jellyfish_name, args.mode, args.out_dir, args.ncpus, args.is_verbose)
    #if run_soap:
    #    execute_soap(args.out_dir, reference_path, args.ncpus, args.is_verbose)
    if run_gff:
        compare_soap_to_gff(args.out_dir, gff_path)
    
